"""
Wake superposition models.
"""

from .ws_linear import WSLinear, WSLinearLocal
from .ws_quadratic import WSQuadratic, WSQuadraticLocal
from .ws_pow import WSPow, WSPowLocal
from .ws_max import WSMax, WSMaxLocal
from .ws_product import WSProduct
from .ti_linear import TILinear
from .ti_quadratic import TIQuadratic
from .ti_pow import TIPow
from .ti_max import TIMax
