"""
Functions for adding turbines to the wind farm.
"""

from .row import add_row
from .grid import add_grid
from .from_json import add_from_json
from .from_csv import add_from_csv
from .from_file import add_from_file
from .from_df import add_from_df
from .from_random import add_random
