from foxes.core import FarmController


class BasicFarmController(FarmController):
    """
    The basic farm controller, running
    all turbine models in their order, without
    conditions.

    :group: models.farm_controllers

    """

    pass
