"""
Algorithms define the main calculation routines.
"""

from .downwind.downwind import Downwind
from .iterative.iterative import Iterative
from .sequential.sequential import Sequential

from . import downwind
from . import iterative
from . import sequential
