from .farm_data_model import FarmDataModel


class FarmModel(FarmDataModel):
    """
    A FarmDataModel that operates on the farm level
    (in contrast to a TurbineModel, which allows
    application to a subset of turbines).

    :group: core

    """

    pass
