"""
Wind deficit wake models.
"""

from .jensen import JensenWake
from .bastankhah14 import Bastankhah2014
from .bastankhah16 import Bastankhah2016Model, Bastankhah2016
from .turbopark import TurbOParkWake, TurbOParkWakeIX
